#undef REQUIRE_EXTENSIONS
#tryinclude <socket>
#tryinclude <curl>
#tryinclude <steamtools>
#tryinclude <SteamWorks>

#define PLUGIN_NAME     "Check FastDL"
#define PLUGIN_VERSION  "1.0"

#define SOCKET_ON()     (GetFeatureStatus(FeatureType_Native, "SocketCreate")                 == FeatureStatus_Available)
#define CURL_ON()       (GetFeatureStatus(FeatureType_Native, "curl_easy_init")               == FeatureStatus_Available)
#define STEAMTOOLS_ON() (GetFeatureStatus(FeatureType_Native, "Steam_CreateHTTPRequest")      == FeatureStatus_Available)
#define STEAMWORKS_ON() (GetFeatureStatus(FeatureType_Native, "SteamWorks_CreateHTTPRequest") == FeatureStatus_Available)

#define MAX_URL_LENGTH  256

new Handle:RedownloadAction = INVALID_HANDLE, 
Handle:Download_URL = INVALID_HANDLE, 
Handle:SaveDirectory = INVALID_HANDLE, 
Handle:MyFileSize = INVALID_HANDLE, 
Handle:CheckTime = INVALID_HANDLE,  // How many seconds to check the file size.
Handle:sv_downloadurl = INVALID_HANDLE,  // Handle variable sv_downloadurl (to optimize)
String:FastDownloadURL[PLATFORM_MAX_PATH],  // To preserve the original link sv_downloadurl
String:CurrentMap[PLATFORM_MAX_PATH];

public Plugin:myinfo = 
{
	name = PLUGIN_NAME, 
	author = "Root", 
	description = "Checks server FastDL", 
	version = PLUGIN_VERSION, 
	url = "http://dodsplugins.net/", 
};


public APLRes:AskPluginLoad2(Handle:myself, bool:late, String:error[], err_max)
{
	#if defined _socket_included
	MarkNativeAsOptional("SocketCreate");
	MarkNativeAsOptional("SocketSetArg");
	MarkNativeAsOptional("SocketSetOption");
	MarkNativeAsOptional("SocketConnect");
	MarkNativeAsOptional("SocketSend");
	#endif
	
	#if defined _cURL_included
	MarkNativeAsOptional("curl_OpenFile");
	MarkNativeAsOptional("curl_slist");
	MarkNativeAsOptional("curl_slist_append");
	MarkNativeAsOptional("curl_easy_init");
	MarkNativeAsOptional("curl_easy_setopt_int_array");
	MarkNativeAsOptional("curl_easy_setopt_handle");
	MarkNativeAsOptional("curl_easy_setopt_string");
	MarkNativeAsOptional("curl_easy_perform_thread");
	MarkNativeAsOptional("curl_easy_strerror");
	#endif
	
	#if defined _steamtools_included
	MarkNativeAsOptional("Steam_CreateHTTPRequest");
	MarkNativeAsOptional("Steam_SetHTTPRequestHeaderValue");
	MarkNativeAsOptional("Steam_SendHTTPRequest");
	MarkNativeAsOptional("Steam_WriteHTTPResponseBody");
	MarkNativeAsOptional("Steam_ReleaseHTTPRequest");
	#endif
	
	#if defined _SteamWorks_Included
	MarkNativeAsOptional("SteamWorks_CreateHTTPRequest");
	MarkNativeAsOptional("SteamWorks_SetHTTPRequestHeaderValue");
	MarkNativeAsOptional("SteamWorks_SetHTTPCallbacks");
	MarkNativeAsOptional("SteamWorks_WriteHTTPResponseBodyToFile");
	MarkNativeAsOptional("SteamWorks_SendHTTPRequest");
	#endif
	return APLRes_Success;
}

public OnPluginStart()
{
	RedownloadAction = CreateConVar("sm_fastdl_refresh", "1", "Specifies the frequency to check fastdl:\n0 - Only when the server starts\n1 - Each time you change the map", FCVAR_PLUGIN, true, 0.0, true, 1.0);
	Download_URL = CreateConVar("sm_fastdl_url", "", "Specify a direct link to the file fastdl\nFor example: \"www.myhost.ru/csgo/maps/de_dust2_se.bsp.bz2\"", FCVAR_PLUGIN);
	SaveDirectory = CreateConVar("sm_fastdl_dir", "", "Specify the folder where to save \"screening \" file\nFor example: \"data/de_dust2_se.bsp.bz2\"", FCVAR_PLUGIN);
	MyFileSize = CreateConVar("sm_fastdl_size", "150", "Specify the estimated size of the test file.\nFastdl shut down if the size will be smaller (in bytes)", FCVAR_PLUGIN, true, 1.0);
	CheckTime = CreateConVar("sm_fastdl_checktime", "30", "Specify how many seconds to check the downloaded file and apply an action to a variable sv_downloadurl", FCVAR_PLUGIN, true, 1.0);
	
	sv_downloadurl = FindConVar("sv_downloadurl");
	
	RegServerCmd("check_fastdl", Command_Download);
	CreateTimer(1.0, Timer_DelayedDownload);
	
	AutoExecConfig();
}

public Action:Timer_DelayedDownload(Handle:timer)
{
	GetConVarString(sv_downloadurl, FastDownloadURL, sizeof(FastDownloadURL)); // Preserve the original link sv_downloadurl
	DownloadFile();
}

public OnConfigsExecuted()
{
	GetCurrentMap(CurrentMap, sizeof(CurrentMap));
	StrCat(CurrentMap, sizeof(CurrentMap), ".bsp.bz2");
	if (GetConVarBool(RedownloadAction))
	{
		DownloadFile();
	}
}

public Action:Command_Download(args)
{
	DownloadFile();
}

DownloadFile()
{
	decl String:mypath[PLATFORM_MAX_PATH], String:url[PLATFORM_MAX_PATH];
	GetConVarString(SaveDirectory, mypath, sizeof(mypath));
	GetConVarString(Download_URL, url, sizeof(url));
	
	if (mypath[0] && url[0])
	{
		BuildPath(PathType:Path_SM, mypath, sizeof(mypath), mypath);
		
		StrCat(mypath, sizeof(mypath), CurrentMap);
		StrCat(url, sizeof(url), CurrentMap);
		DeleteFile(mypath);
		
		if (CanTestFeatures() && SOCKET_ON())
		{
			#if defined _socket_included
			Download_Socket(url, mypath);
			#endif
		}
		
		else if (CanTestFeatures() && CURL_ON())
		{
			#if defined _cURL_included
			Download_cURL(url, mypath);
			#endif
		}
		
		else if (CanTestFeatures() && STEAMTOOLS_ON())
		{
			#if defined _steamtools_included
			Download_SteamTools(url, mypath);
			#endif
		}
		
		else if (CanTestFeatures() && STEAMWORKS_ON())
		{
			#if defined _SteamWorks_Included
			Download_SteamWorks(url, mypath);
			#endif
		}
		
		else SetFailState("This plugin requires either the Socket, cURL, SteamTools or SteamWorks extension to work!");
		
		new Handle:data = INVALID_HANDLE;
		CreateDataTimer(GetConVarFloat(CheckTime), Timer_CheckFile, data);
		WritePackString(data, mypath);
	}
}

public Action:Timer_CheckFile(Handle:timer, any:data)
{
	if (data == INVALID_HANDLE)
	{
		LogError("Invalid data timer!");
		return Plugin_Stop;
	}
	
	ResetPack(data);
	
	new String:mypath[PLATFORM_MAX_PATH], size;
	ReadPackString(data, mypath, sizeof(mypath));
	size = FileSize(mypath);
	
	if (size < GetConVarInt(MyFileSize))
	{
		LogMessage("The size of test file is %i, but should be more than %i. Disable FastDL", size, GetConVarInt(MyFileSize));
		SetConVarString(sv_downloadurl, "");
	}
	else
	{
		SetConVarString(sv_downloadurl, FastDownloadURL);
	}
	
	// We remove the test file
	DeleteFile(mypath);
	
	return Plugin_Stop;
}

#if defined _socket_included
Download_Socket(const String:url[], const String:folder[])
{
	new Handle:file = OpenFile(folder, "wb");
	if (file == INVALID_HANDLE)
	{
		decl String:error[PLATFORM_MAX_PATH];
		FormatEx(error, sizeof(error), "Unable to write data in \"%s\"", folder);
	}
	
	decl String:hostname[64], String:location[128], String:filename[64], String:request[MAX_URL_LENGTH + 128];
	ParseURL(url, hostname, sizeof(hostname), location, sizeof(location), filename, sizeof(filename));
	FormatEx(request, sizeof(request), "GET %s/%s HTTP/1.0\r\nHost: %s\r\nUser-agent: plugin\r\nConnection: close\r\nPragma: no-cache\r\nCache-Control: no-cache\r\n\r\n", location, filename, hostname);
	
	new Handle:hDLPack = CreateDataPack();
	WritePackCell(hDLPack, 0);
	WritePackCell(hDLPack, _:file);
	WritePackString(hDLPack, request);
	
	new Handle:socket = SocketCreate(SOCKET_TCP, OnSocketError);
	SocketSetArg(socket, hDLPack);
	SocketSetOption(socket, ConcatenateCallbacks, 4096);
	SocketConnect(socket, OnSocketConnected, OnSocketReceive, OnSocketDisconnected, hostname, 80);
}

public OnSocketConnected(Handle:socket, any:hDLPack)
{
	decl String:request[MAX_URL_LENGTH + 128];
	SetPackPosition(hDLPack, 16);
	ReadPackString(hDLPack, request, sizeof(request));
	
	SocketSend(socket, request);
}

public OnSocketReceive(Handle:socket, String:data[], const size, any:hDLPack)
{
	new idx = 0;
	
	SetPackPosition(hDLPack, 0);
	new bool:bParsedHeader = bool:ReadPackCell(hDLPack);
	
	if (!bParsedHeader)
	{
		if ((idx = StrContains(data, "\r\n\r\n")) == -1)
			idx = 0;
		else
			idx += 4;
		
		SetPackPosition(hDLPack, 0);
		WritePackCell(hDLPack, 1);
	}
	
	SetPackPosition(hDLPack, 8);
	new Handle:file = Handle:ReadPackCell(hDLPack);
	
	while (idx < size)
	{
		WriteFileCell(file, data[idx++], 1);
	}
}

public OnSocketDisconnected(Handle:socket, any:hDLPack)
{
	SetPackPosition(hDLPack, 8);
	CloseHandle(Handle:ReadPackCell(hDLPack));
	CloseHandle(hDLPack);
	CloseHandle(socket);
}

public OnSocketError(Handle:socket, const errorType, const errorNum, any:hDLPack)
{
	SetPackPosition(hDLPack, 8);
	CloseHandle(Handle:ReadPackCell(hDLPack));
	CloseHandle(hDLPack);
	CloseHandle(socket);
	
	decl String:error[256];
	FormatEx(error, sizeof(error), "Socket error: %d (Error code %d)", errorType, errorNum);
}

ParseURL(const String:url[], String:host[], maxHost, String:location[], maxLoc, String:filename[], maxName)
{
	new idx = StrContains(url, "://");
	idx = (idx != -1) ? idx + 3 : 0;
	
	decl String:dirs[16][64];
	new total = ExplodeString(url[idx], "/", dirs, sizeof(dirs), sizeof(dirs[]));
	
	FormatEx(host, maxHost, "%s", dirs[0]);
	
	location[0] = '\0';
	for (new i = 1; i < total - 1; i++)
	{
		FormatEx(location, maxLoc, "%s/%s", location, dirs[i]);
	}
	
	FormatEx(filename, maxName, "%s", dirs[total - 1]);
}
#endif

#if defined _cURL_included
Download_cURL(const String:url[], const String:folder[])
{
	decl String:sURL[MAX_URL_LENGTH];
	PrefixURL(sURL, sizeof(sURL), url);
	
	new Handle:file = curl_OpenFile(folder, "wb");
	if (file == INVALID_HANDLE)
	{
		decl String:error[PLATFORM_MAX_PATH];
		FormatEx(error, sizeof(error), "Unable to write data in \"%s\"", folder);
	}
	
	new CURL_Default_opt[][2] =  {
		{ _:CURLOPT_NOSIGNAL, 1 }, 
		{ _:CURLOPT_NOPROGRESS, 1 }, 
		{ _:CURLOPT_TIMEOUT, 30 }, 
		{ _:CURLOPT_CONNECTTIMEOUT, 60 }, 
		{ _:CURLOPT_VERBOSE, 0 }
	};
	
	new Handle:headers = curl_slist();
	curl_slist_append(headers, "User-agent: plugin");
	curl_slist_append(headers, "Pragma: no-cache");
	curl_slist_append(headers, "Cache-Control: no-cache");
	
	new Handle:hDLPack = CreateDataPack();
	WritePackCell(hDLPack, _:file);
	WritePackCell(hDLPack, _:headers);
	
	new Handle:curl = curl_easy_init();
	curl_easy_setopt_int_array(curl, CURL_Default_opt, sizeof(CURL_Default_opt));
	curl_easy_setopt_handle(curl, CURLOPT_WRITEDATA, file);
	curl_easy_setopt_string(curl, CURLOPT_URL, url);
	curl_easy_setopt_handle(curl, CURLOPT_HTTPHEADER, headers);
	curl_easy_perform_thread(curl, OnCurlComplete, hDLPack);
}

public OnCurlComplete(Handle:curl, CURLcode:code, any:hDLPack)
{
	ResetPack(hDLPack);
	CloseHandle(Handle:ReadPackCell(hDLPack));
	CloseHandle(Handle:ReadPackCell(hDLPack));
	CloseHandle(hDLPack);
	CloseHandle(curl);
	
	if (code != CURLE_OK)
	{
		decl String:error[PLATFORM_MAX_PATH];
		curl_easy_strerror(code, error, sizeof(error));
		FormatEx(error, sizeof(error), "cURL error: %s", error);
	}
}
#endif

#if defined _steamtools_included
Download_SteamTools(const String:url[], const String:folder[])
{
	new Handle:file = OpenFile(folder, "wb");
	if (file == INVALID_HANDLE)
	{
		decl String:error[PLATFORM_MAX_PATH];
		FormatEx(error, sizeof(error), "Unable to write data in \"%s\"", folder);
	}
	
	decl String:sURL[MAX_URL_LENGTH];
	PrefixURL(sURL, sizeof(sURL), url);
	
	new Handle:hDLPack = CreateDataPack();
	WritePackString(hDLPack, folder);
	
	new HTTPRequestHandle:hRequest = Steam_CreateHTTPRequest(HTTPMethod_GET, sURL);
	Steam_SetHTTPRequestHeaderValue(hRequest, "User-agent", "plugin");
	Steam_SetHTTPRequestHeaderValue(hRequest, "Pragma", "no-cache");
	Steam_SetHTTPRequestHeaderValue(hRequest, "Cache-Control", "no-cache");
	Steam_SendHTTPRequest(hRequest, OnSteamHTTPComplete, hDLPack);
}

public OnSteamHTTPComplete(HTTPRequestHandle:HTTPRequest, bool:requestSuccessful, HTTPStatusCode:statusCode, any:hDLPack)
{
	decl String:file[PLATFORM_MAX_PATH];
	ResetPack(hDLPack);
	ReadPackString(hDLPack, file, sizeof(file));
	CloseHandle(hDLPack);
	
	if (requestSuccessful && statusCode == HTTPStatusCode_OK)
	{
		Steam_WriteHTTPResponseBody(HTTPRequest, file);
	}
	
	else
	{
		decl String:error[PLATFORM_MAX_PATH];
		FormatEx(error, sizeof(error), "SteamTools error (status code %i). Request successful: %s", _:statusCode, requestSuccessful ? "Yes" : "No");
	}
	
	Steam_ReleaseHTTPRequest(HTTPRequest);
}
#endif

#if defined _SteamWorks_Included
Download_SteamWorks(const String:url[], const String:folder[])
{
	new Handle:file = OpenFile(folder, "wb");
	if (file == INVALID_HANDLE)
	{
		decl String:error[PLATFORM_MAX_PATH];
		FormatEx(error, sizeof(error), "Unable to write data in \"%s\"", folder);
	}
	
	decl String:sURL[MAX_URL_LENGTH];
	PrefixURL(sURL, sizeof(sURL), url);
	
	new Handle:hRequest = SteamWorks_CreateHTTPRequest(k_EHTTPMethodGET, sURL);
	SteamWorks_SetHTTPRequestHeaderValue(hRequest, "User-agent", "plugin");
	SteamWorks_SetHTTPRequestHeaderValue(hRequest, "Pragma", "no-cache");
	SteamWorks_SetHTTPRequestHeaderValue(hRequest, "Cache-Control", "no-cache");
	SteamWorks_SetHTTPCallbacks(hRequest, OnSWHTTPComplete);
	SteamWorks_SendHTTPRequest(hRequest);
}

public OnSWHTTPComplete(Handle:HTTPRequest, bool:bFailure, bool:bRequestSuccessful, EHTTPStatusCode:eStatusCode, any:data1)
{
	decl String:mypath[PLATFORM_MAX_PATH];
	GetConVarString(SaveDirectory, mypath, sizeof(mypath));
	BuildPath(PathType:Path_SM, mypath, sizeof(mypath), mypath);
	
	if (bRequestSuccessful && eStatusCode == k_EHTTPStatusCode200OK)
	{
		SteamWorks_WriteHTTPResponseBodyToFile(HTTPRequest, mypath);
	}
	
	else
	{
		decl String:error[PLATFORM_MAX_PATH];
		FormatEx(error, sizeof(error), "SteamWorks error (status code %i). Request successful: %s", _:eStatusCode, bRequestSuccessful ? "Yes" : "No");
	}
	
	CloseHandle(HTTPRequest);
}
#endif

#if defined _cURL_included || defined _steamtools_included || defined _SteamWorks_Included
PrefixURL(String:buffer[], maxlength, const String:url[])
{
	if (strncmp(url, "http://", 7) != 0)
	{
		FormatEx(buffer, maxlength, "http://%s", url);
	}
	
	else strcopy(buffer, maxlength, url);
}
#endif